@@ -1,5 +1,5 @@
 /*
- *Copyright (C) 2004-2006 Qball Cow <Qball@qballcow.nl>
+ *Copyright (C) 2004-2007 Qball Cow <Qball@qballcow.nl>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -709,3 +709,321 @@ MpdData * mpd_database_get_directory(MpdObj *mi,char *path)
 	return mpd_data_get_first(data);
 }
 
+MpdData *mpd_database_get_playlist_content(MpdObj *mi,char *playlist)
+{
+	MpdData *data = NULL;
+	mpd_InfoEntity *ent = NULL;
+	if(!mpd_check_connected(mi))
+	{
+		debug_printf(DEBUG_WARNING,"not connected\n");
+		return NULL;
+	}
+	if(!mpd_server_check_version(mi, 0,12,0))
+	{
+		debug_printf(DEBUG_WARNING, "only works with mpd higher then 0.12.0");
+		return NULL;
+	}
+	if(mpd_server_check_command_allowed(mi, "listplaylistinfo") != MPD_SERVER_COMMAND_ALLOWED)
+	{
+		debug_printf(DEBUG_WARNING, "Listing playlist content not supported or allowed");
+		return NULL;
+	}
+	if(mpd_lock_conn(mi))
+	{
+		debug_printf(DEBUG_WARNING,"lock failed\n");
+		return NULL;
+	}
+	mpd_sendListPlaylistInfoCommand(mi->connection, playlist);
+	while (( ent = mpd_getNextInfoEntity(mi->connection)) != NULL)
+	{
+		data = mpd_new_data_struct_append( data );
+		if(ent->type == MPD_INFO_ENTITY_TYPE_DIRECTORY)
+		{
+			data->type = MPD_DATA_TYPE_DIRECTORY;
+			data->directory = ent->info.directory->path;
+			ent->info.directory->path = NULL;
+		}
+		else if (ent->type == MPD_INFO_ENTITY_TYPE_SONG)
+		{
+			data->type = MPD_DATA_TYPE_SONG;
+			data->song = ent->info.song;
+			ent->info.song = NULL;
+		}
+		else if (ent->type == MPD_INFO_ENTITY_TYPE_PLAYLISTFILE)
+		{
+			data->type = MPD_DATA_TYPE_PLAYLIST;
+			data->playlist = ent->info.playlistFile->path;
+			ent->info.playlistFile->path = NULL;
+		}
+
+		mpd_freeInfoEntity(ent);
+	}
+	mpd_finishCommand(mi->connection);
+
+	/* unlock */
+	mpd_unlock_conn(mi);
+	if(data == NULL)
+	{
+		return NULL;
+	}
+	return mpd_data_get_first(data);
+}
+/**
+ * @param mi A #MpdObj
+ * @param path an Path to a file
+ *
+ * Grabs the song info for a single file. Make sure you pass an url to a song
+ * and not a directory, that might result in strange behauviour.
+ *
+ * @returns a #mpd_Song
+ */
+mpd_Song * mpd_database_get_fileinfo(MpdObj *mi,const char *path)
+{
+	mpd_Song *song = NULL;
+	mpd_InfoEntity *ent = NULL;
+	/*
+	 * Check path for availibility and length
+	 */
+	if(path == NULL || path[0] == '\0')
+	{
+		debug_printf(DEBUG_ERROR, "path == NULL || strlen(path) == 0");
+		return NULL;
+	}
+	if(!mpd_check_connected(mi))
+	{
+		debug_printf(DEBUG_ERROR, "Not Connected\n");
+		return NULL;
+	}
+	/* lock, so we can work on mi->connection */
+	if(mpd_lock_conn(mi) != MPD_OK)
+	{
+		debug_printf(DEBUG_ERROR, "Failed to lock connection");
+		return NULL;
+	}
+	/* send the request */
+	mpd_sendListallInfoCommand(mi->connection, path);
+	/* get the first (and only) result */
+	ent = mpd_getNextInfoEntity(mi->connection);
+	/* finish and clean up libmpdclient */
+	mpd_finishCommand(mi->connection);
+	/* unlock again */
+	if(mpd_unlock_conn(mi))
+	{
+		if(ent) mpd_freeInfoEntity(ent);
+		debug_printf(DEBUG_ERROR, "Failed to unlock");
+		return NULL;
+	}
+
+	if(ent == NULL)
+	{
+		debug_printf(DEBUG_ERROR, "Failed to grab song from mpd\n");
+		return NULL;
+	}
+
+	if(ent->type != MPD_INFO_ENTITY_TYPE_SONG)
+	{
+		mpd_freeInfoEntity(ent);
+		debug_printf(DEBUG_ERROR, "Failed to grab correct song type from mpd, path might not be a file\n");
+		return NULL;
+	}
+	/* get the song */
+	song = ent->info.song;
+	/* remove reference to song from the entity */
+	ent->info.song = NULL;
+	/* free the entity */
+	mpd_freeInfoEntity(ent);
+
+	return song;
+}
+
+
+void mpd_database_search_field_start(MpdObj *mi, mpd_TagItems field)
+{
+	/*
+	 * Check argument
+	 */
+	if(mi == NULL || field >= MPD_TAG_NUM_OF_ITEM_TYPES || field < 0) 
+	{
+		debug_printf(DEBUG_ERROR, "Argument error");
+		return ;
+	}
+	if(!mpd_check_connected(mi))
+	{
+		debug_printf(DEBUG_ERROR, "Not Connected\n");
+		return ;
+	}
+	if(!mpd_server_check_version(mi, 0,12,0))
+	{
+		debug_printf(DEBUG_ERROR, "Advanced field list requires mpd 0.12.0 or higher");
+		return ;
+	}
+	/* lock, so we can work on mi->connection */
+	if(mpd_lock_conn(mi) != MPD_OK)
+	{
+		debug_printf(DEBUG_ERROR, "Failed to lock connection");
+		return ;
+	}
+	mpd_startFieldSearch(mi->connection, field);
+	/* Set search type */
+	mi->search_type = MPD_SEARCH_TYPE_LIST;
+	mi->search_field = field;
+	/* unlock, let the error handler handle any possible error.
+	 */
+	mpd_unlock_conn(mi);
+	return;
+}
+
+
+
+
+
+
+
+/**
+ * @param mi A #MpdObj
+ * @param exact a boolean indicating if the search is fuzzy or exact
+ *
+ * Starts a search, you can add "constraints" by calling mpd_database_search_add_constraint
+ * 
+ * This function requires mpd 0.12.0 or higher 
+ */
+
+void mpd_database_search_start(MpdObj *mi, int exact)
+{
+	/*
+	 * Check argument
+	 */
+	if(mi == NULL || exact > 1 || exact < 0) 
+	{
+		debug_printf(DEBUG_ERROR, "Argument error");
+		return ;
+	}
+	if(!mpd_check_connected(mi))
+	{
+		debug_printf(DEBUG_ERROR, "Not Connected\n");
+		return ;
+	}
+	if(!mpd_server_check_version(mi, 0,12,0))
+	{
+		debug_printf(DEBUG_ERROR, "Advanced search requires mpd 0.12.0 or higher");
+		return ;
+	}
+	/* lock, so we can work on mi->connection */
+	if(mpd_lock_conn(mi) != MPD_OK)
+	{
+		debug_printf(DEBUG_ERROR, "Failed to lock connection");
+		return ;
+	}
+	mpd_startSearch(mi->connection, exact);
+	/* Set search type */
+	mi->search_type = (exact)? MPD_SEARCH_TYPE_FIND:MPD_SEARCH_TYPE_SEARCH;
+	/* unlock, let the error handler handle any possible error.
+	 */
+	mpd_unlock_conn(mi);
+	return;
+}
+/**
+ * @param mi A #MpdObj
+ * @param field A #mpd_TagItems
+ *
+ * Adds a constraint to the search 
+ */
+void mpd_database_search_add_constraint(MpdObj *mi, mpd_TagItems field, char *value)
+{
+	if(mi == NULL || value == NULL || value[0] == '\0')
+	{
+		debug_printf(DEBUG_ERROR,"Failed to parse arguments");
+		return;
+	}
+	if(mi->search_type == MPD_SEARCH_TYPE_NONE)
+	{
+		debug_printf(DEBUG_ERROR, "No search to constraint");
+		return;
+	}
+	if(!mpd_check_connected(mi))
+	{
+		debug_printf(DEBUG_ERROR, "Not Connected\n");
+		return ;
+	}
+	if(!mpd_server_check_version(mi, 0,12,0))
+	{
+		debug_printf(DEBUG_ERROR, "Advanced search requires mpd 0.12.0 or higher");
+		return ;
+	}
+	/* lock, so we can work on mi->connection */
+	if(mpd_lock_conn(mi) != MPD_OK)
+	{
+		debug_printf(DEBUG_ERROR, "Failed to lock connection");
+		return ;
+	}
+	mpd_addConstraintSearch(mi->connection, field, value);
+	/* unlock, let the error handler handle any possible error.
+	 */
+	mpd_unlock_conn(mi);
+	return;
+}
+
+MpdData * mpd_database_search_commit(MpdObj *mi)
+{
+	mpd_InfoEntity *ent = NULL;
+	MpdData *data = NULL;
+	if(!mpd_check_connected(mi))
+	{
+		debug_printf(DEBUG_WARNING,"not connected\n");
+		return NULL;
+	}
+	if(mi->search_type == MPD_SEARCH_TYPE_NONE)
+	{
+		debug_printf(DEBUG_ERROR, "no search in progress to commit");
+		return NULL;
+	}
+	if(mpd_lock_conn(mi))
+	{
+		debug_printf(DEBUG_ERROR,"lock failed\n");
+		return NULL;
+	}
+	mpd_commitSearch(mi->connection);
+	if(mi->search_type == MPD_SEARCH_TYPE_LIST)
+	{
+		char *string = NULL;
+		while((string = mpd_getNextTag(mi->connection, mi->search_field)))
+		{
+			data = mpd_new_data_struct_append(data);
+			data->type = MPD_DATA_TYPE_TAG;
+			data->tag_type = mi->search_field;
+			data->tag = string;
+		}
+	}
+	else
+	{
+		while (( ent = mpd_getNextInfoEntity(mi->connection)) != NULL)
+		{
+			if(ent->type == MPD_INFO_ENTITY_TYPE_SONG)
+			{
+				data = mpd_new_data_struct_append(data);
+				data->type = MPD_DATA_TYPE_SONG;
+				data->song = ent->info.song;
+				ent->info.song = NULL;
+			}
+			mpd_freeInfoEntity(ent);
+		}
+	}
+	mpd_finishCommand(mi->connection);
+	/*
+	 * reset search type
+	 */
+	mi->search_type = MPD_SEARCH_TYPE_NONE;
+	mi->search_field = MPD_TAG_ITEM_ARTIST;
+	/* unlock */
+	if(mpd_unlock_conn(mi))
+	{
+		debug_printf(DEBUG_ERROR, "Failed to unlock connection");
+		if(data)mpd_data_free(data);
+		return NULL;
+	}
+	if(data == NULL)
+	{
+		return NULL;
+	}
+	return mpd_data_get_first(data);
+}
