@@ -1,5 +1,5 @@
 /*
- *Copyright (C) 2004-2006 Qball Cow <Qball@qballcow.nl>
+ *Copyright (C) 2004-2007 Qball Cow <Qball@qballcow.nl>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -320,6 +320,38 @@ int mpd_stats_get_playtime(MpdObj *mi)
 	}
 	return mi->stats->playTime;
 }
+
+int mpd_stats_get_db_playtime(MpdObj *mi)
+{
+	if(mi == NULL)
+	{
+		debug_printf(DEBUG_ERROR, "failed to check mi == NULL\n");
+		return MPD_ARGS_ERROR;
+	}
+	if(mpd_stats_check(mi) != MPD_OK)
+	{
+		debug_printf(DEBUG_WARNING,"Failed to get stats\n");
+		return MPD_STATS_FAILED;
+	}
+	return mi->stats->dbPlayTime;
+}
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
 int mpd_status_get_volume(MpdObj *mi)
 {
 	if(mi == NULL)
