@@ -1,5 +1,5 @@
 /*
- *Copyright (C) 2004-2006 Qball Cow <Qball@qballcow.nl>
+ *Copyright (C) 2004-2007 Qball Cow <Qball@qballcow.nl>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -30,7 +30,10 @@
 #include "libmpd.h"
 #include "libmpd-internal.h"
 
+static void mpd_free_queue_ob(MpdObj *mi);
+static void mpd_server_free_commands(MpdObj *mi);
 
+char *libmpd_version = LIBMPD_VERSION;
 #ifndef HAVE_STRNDUP
 /**
  * Not every platfarm has strndup, so here we have a nice little custom implementation
@@ -135,6 +138,10 @@ static MpdObj * mpd_create()
 
 	/* set queue to NULL */
 	mi->queue = NULL;
+	/* search stuff */
+	mi->search_type = MPD_SEARCH_TYPE_NONE;
+	/* no need to initialize, but set it to anything anyway*/
+	mi->search_field = MPD_TAG_ITEM_ARTIST;
 
 	/* commands */
 	mi->commands = NULL;
@@ -147,8 +154,8 @@ void mpd_free(MpdObj *mi)
 	if(mi->connected)
 	{
 		/* disconnect */
-		mpd_disconnect(mi);
 		debug_printf(DEBUG_WARNING, "Connection still running, disconnecting\n");
+		mpd_disconnect(mi);
 	}
 	if(mi->hostname)
 	{
@@ -179,6 +186,8 @@ void mpd_free(MpdObj *mi)
 	{
 		mpd_freeSong(mi->CurrentSong);
 	}
+	mpd_free_queue_ob(mi);
+	mpd_server_free_commands(mi);		
 	free(mi);
 }
 
@@ -294,6 +303,16 @@ MpdObj *mpd_new(char *hostname,  int port, char *password)
 	return mi;
 }
 
+
+const char * mpd_get_hostname(MpdObj *mi)
+{
+	if(mi == NULL)
+	{
+		return NULL;
+	}
+	return mi->hostname;
+}
+
 int mpd_set_hostname(MpdObj *mi, char *hostname)
 {
 	if(mi == NULL)
@@ -332,7 +351,7 @@ int mpd_set_password(MpdObj *mi, char *password)
 int mpd_send_password(MpdObj *mi)
 {
 	if(!mi) return MPD_ARGS_ERROR;
-	if(mi->password && mpd_check_connected(mi))
+	if(mi->password && mpd_check_connected(mi) && strlen(mi->password))
 	{
 		if(mpd_lock_conn(mi))
 		{
@@ -411,6 +430,17 @@ static void mpd_server_free_commands(MpdObj *mi)
 	}
 }
 
+char *mpd_server_get_version(MpdObj *mi)
+{
+	char *retval = NULL;
+	if(!mi || !mpd_check_connected(mi))
+		return NULL;
+	retval = malloc(10*sizeof(char));
+	snprintf(retval,10,"%i.%i.%i", mi->connection->version[0], mi->connection->version[1], mi->connection->version[2]);
+	/* always make sure the string is terminated */
+	retval[9] = '\0';
+	return retval;
+}
 
 int mpd_server_get_allowed_commands(MpdObj *mi)
 {
@@ -510,21 +540,25 @@ int mpd_disconnect(MpdObj *mi)
 	mi->CurrentState.channels = 0; 
 	mi->CurrentState.bits = 0;
 
+	/* search stuff */
+	mi->search_type = MPD_SEARCH_TYPE_NONE;
+	/* no need to initialize, but set it to anything anyway*/
+	mi->search_field = MPD_TAG_ITEM_ARTIST;
+
+
+
+	
 	memcpy(&(mi->OldState), &(mi->CurrentState) , sizeof(MpdServerState));
+
+	mpd_free_queue_ob(mi);
+	mpd_server_free_commands(mi);	
 	/*don't reset errors */
-
-
 	if(mi->the_connection_changed_callback != NULL)
 	{
 		mi->the_connection_changed_callback( mi, FALSE, mi->the_connection_changed_signal_userdata );
 	}
-	/* deprecated */
-	/*	if(mi->disconnect != NULL)
-		{
-		mi->disconnect(mi, mi->disconnect_pointer);
-		}
-		*/
-	mpd_server_free_commands(mi);
+	
+	debug_printf(DEBUG_INFO, "Disconnect completed\n");
 	return MPD_OK;
 }
 
@@ -572,6 +606,11 @@ int mpd_connect(MpdObj *mi)
 		/* TODO: make seperate error message? */
 		return MPD_NOT_CONNECTED;
 	}
+	if(mpd_check_error(mi) != MPD_OK)
+	{
+		/* TODO: make seperate error message? */
+		return MPD_NOT_CONNECTED;
+	}
 
 	/* set connected state */
 	mi->connected = TRUE;
@@ -847,6 +886,10 @@ void mpd_data_free(MpdData *data)
 				if(data_real->song) mpd_freeSong(data_real->song);
 			} else if (data_real->type == MPD_DATA_TYPE_OUTPUT_DEV) {
 				mpd_freeOutputElement(data_real->output_dev);
+			} else if(data_real->type == MPD_DATA_TYPE_DIRECTORY) {
+				if(data_real->directory)free(data_real->directory);
+			} else if(data_real->type == MPD_DATA_TYPE_PLAYLIST) {
+				if(data_real->playlist)free(data_real->playlist);				
 			} else {
 				free((void*)(data_real->tag));
 			}
@@ -869,7 +912,7 @@ static void mpd_free_queue_ob(MpdObj *mi)
 	}
 	if(mi->queue == NULL)
 	{
-		debug_printf(DEBUG_ERROR, "mi->queue != NULL failed");
+		debug_printf(DEBUG_INFO, "mi->queue != NULL failed");
 		return;
 	}	
 	mi->queue = mi->queue->first;
