@@ -1,5 +1,5 @@
 /* libmpdclient
-   (c)2003-2004 by Warren Dukes (shank@mercury.chem.pitt.edu)
+   (c)2003-2007 by Warren Dukes (warren.dukes@gmail.com)
    This project's homepage is: http://www.musicpd.org
 
    Redistribution and use in source and binary forms, with or without
@@ -28,49 +28,211 @@
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-
 */
 
-
 #include "libmpdclient.h"
 
 #include <errno.h>
+#include <ctype.h>
 #include <sys/types.h>
 #include <stdio.h>
 #include <sys/param.h>
-
-
 #include <string.h>
-#include <strings.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <fcntl.h>
-#include <stdarg.h>
-
 
 #ifdef WIN32
-#include <ws2tcpip.h>
-#include <winsock.h>
+#  include <ws2tcpip.h>
+#  include <winsock.h>
 #else
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#include <sys/socket.h>
-#include <netdb.h>
-#endif
-
-
-#ifndef MPD_NO_IPV6
-#ifdef AF_INET6
-#define MPD_HAVE_IPV6
-#endif
+#  include <netinet/in.h>
+#  include <arpa/inet.h>
+#  include <sys/socket.h>
+#  include <netdb.h>
 #endif
 
 #ifndef MSG_DONTWAIT
-#define MSG_DONTWAIT 0
+#  define MSG_DONTWAIT 0
 #endif
 
-#define COMMAND_LIST	1
-#define COMMAND_LIST_OK	2
+#ifndef MPD_NO_GAI
+#  ifdef AI_ADDRCONFIG
+#    define MPD_HAVE_GAI
+#  endif
+#endif
+
+#define COMMAND_LIST    1
+#define COMMAND_LIST_OK 2
+
+#ifdef WIN32
+#  define SELECT_ERRNO_IGNORE   (errno == WSAEINTR || errno == WSAEINPROGRESS)
+#  define SENDRECV_ERRNO_IGNORE SELECT_ERRNO_IGNORE
+#else
+#  define SELECT_ERRNO_IGNORE   (errno == EINTR)
+#  define SENDRECV_ERRNO_IGNORE (errno == EINTR || errno == EAGAIN)
+#  define winsock_dll_error(c)  0
+#  define closesocket(s)        close(s)
+#  define WSACleanup()          do { /* nothing */ } while (0)
+#endif
+
+#ifdef WIN32
+static int winsock_dll_error(mpd_Connection *connection)
+{
+	WSADATA wsaData;
+	if ((WSAStartup(MAKEWORD(2, 2), &wsaData)) != 0 ||
+			LOBYTE(wsaData.wVersion) != 2 ||
+			HIBYTE(wsaData.wVersion) != 2 ) {
+		strcpy(connection->errorStr,
+		       "Could not find usable WinSock DLL.");
+		connection->error = MPD_ERROR_SYSTEM;
+		return 1;
+	}
+	return 0;
+}
+
+static int do_connect_fail(mpd_Connection *connection,
+                           const struct sockaddr *serv_addr, int addrlen)
+{
+	int iMode = 1; /* 0 = blocking, else non-blocking */
+	ioctlsocket(connection->sock, FIONBIO, (u_long FAR*) &iMode);
+	return (connect(connection->sock,serv_addr,addrlen) == SOCKET_ERROR
+			&& WSAGetLastError() != WSAEWOULDBLOCK);
+}
+#else /* !WIN32 (sane operating systems) */
+static int do_connect_fail(mpd_Connection *connection,
+                           const struct sockaddr *serv_addr, int addrlen)
+{
+	int flags = fcntl(connection->sock, F_GETFL, 0);
+	fcntl(connection->sock, F_SETFL, flags | O_NONBLOCK);
+	return (connect(connection->sock,serv_addr,addrlen)<0 &&
+				errno!=EINPROGRESS);
+}
+#endif /* !WIN32 */
+
+#ifdef MPD_HAVE_GAI
+static int mpd_connect(mpd_Connection * connection, const char * host, int port,
+                       float timeout)
+{
+	int error;
+	char service[20];
+	struct addrinfo hints;
+	struct addrinfo *res = NULL;
+	struct addrinfo *addrinfo = NULL;
+
+	/**
+	 * Setup hints
+	 */
+	hints.ai_flags     = AI_ADDRCONFIG;
+	hints.ai_family    = PF_UNSPEC;
+	hints.ai_socktype  = SOCK_STREAM;
+	hints.ai_protocol  = IPPROTO_TCP;
+	hints.ai_addrlen   = 0;
+	hints.ai_addr      = NULL;
+	hints.ai_canonname = NULL;
+	hints.ai_next      = NULL;
+
+	snprintf(service, sizeof(service), "%d", port);
+
+	error = getaddrinfo(host, service, &hints, &addrinfo);
+
+	if (error) {
+		snprintf(connection->errorStr,MPD_ERRORSTR_MAX_LENGTH,
+				"host \"%s\" not found: %s",host, gai_strerror(error));
+		connection->error = MPD_ERROR_UNKHOST;
+		return -1;
+	}
+
+	for (res = addrinfo; res; res = res->ai_next) {
+		/* create socket */
+		connection->sock = socket(res->ai_family, SOCK_STREAM, res->ai_protocol);
+		if (connection->sock < 0) {
+			snprintf(connection->errorStr, MPD_ERRORSTR_MAX_LENGTH,
+			         "problems creating socket: %s",
+			         strerror(errno));
+			connection->error = MPD_ERROR_SYSTEM;
+			freeaddrinfo(addrinfo);
+			return -1;
+		}
+
+		mpd_setConnectionTimeout(connection,timeout);
+
+		/* connect stuff */
+ 		if (do_connect_fail(connection, res->ai_addr, res->ai_addrlen)) {
+ 			/* try the next address family */
+ 			closesocket(connection->sock);
+ 			connection->sock = -1;
+ 			continue;
+		}
+	}
+	freeaddrinfo(addrinfo);
+
+	if (connection->sock < 0) {
+		snprintf(connection->errorStr,MPD_ERRORSTR_MAX_LENGTH,
+				"problems connecting to \"%s\" on port"
+				" %i: %s",host,port, strerror(errno));
+		connection->error = MPD_ERROR_CONNPORT;
+
+		return -1;
+	}
+
+	return 0;
+}
+#else /* !MPD_HAVE_GAI */
+static int mpd_connect(mpd_Connection * connection, const char * host, int port,
+                       float timeout)
+{
+	struct hostent * he;
+	struct sockaddr * dest;
+	int destlen;
+	struct sockaddr_in sin;
+
+	if(!(he=gethostbyname(host))) {
+		snprintf(connection->errorStr,MPD_ERRORSTR_MAX_LENGTH,
+				"host \"%s\" not found",host);
+		connection->error = MPD_ERROR_UNKHOST;
+		return -1;
+	}
+
+	memset(&sin,0,sizeof(struct sockaddr_in));
+	/*dest.sin_family = he->h_addrtype;*/
+	sin.sin_family = AF_INET;
+	sin.sin_port = htons(port);
+
+	switch(he->h_addrtype) {
+	case AF_INET:
+		memcpy((char *)&sin.sin_addr.s_addr,(char *)he->h_addr,
+				he->h_length);
+		dest = (struct sockaddr *)&sin;
+		destlen = sizeof(struct sockaddr_in);
+		break;
+	default:
+		strcpy(connection->errorStr,"address type is not IPv4");
+		connection->error = MPD_ERROR_SYSTEM;
+		return -1;
+		break;
+	}
+
+	if((connection->sock = socket(dest->sa_family,SOCK_STREAM,0))<0) {
+		strcpy(connection->errorStr,"problems creating socket");
+		connection->error = MPD_ERROR_SYSTEM;
+		return -1;
+	}
+
+	mpd_setConnectionTimeout(connection,timeout);
+
+	/* connect stuff */
+	if (do_connect_fail(connection, dest, destlen)) {
+		snprintf(connection->errorStr,MPD_ERRORSTR_MAX_LENGTH,
+				"problems connecting to \"%s\" on port"
+				" %i",host,port);
+		connection->error = MPD_ERROR_CONNPORT;
+		return -1;
+	}
+
+	return 0;
+}
+#endif /* !MPD_HAVE_GAI */
 
 char * mpdTagItemKeys[MPD_TAG_NUM_OF_ITEM_TYPES] =
 {
@@ -84,36 +246,17 @@ char * mpdTagItemKeys[MPD_TAG_NUM_OF_ITEM_TYPES] =
 	"Composer",
 	"Performer",
 	"Comment",
-	"filename"
+	"Disc",
+	"Filename",
+	"Any"
 };
 
-
-#ifdef MPD_HAVE_IPV6        
-static int mpd_ipv6Supported() {
-	int s;          
-	s = socket(AF_INET6,SOCK_STREAM,0);
-	if(s == -1) return 0;
-	close(s);       
-	return 1;                       
-}                       
-#endif  
-
 static char * mpd_sanitizeArg(const char * arg) {
 	size_t i;
 	char * ret;
 	register const char *c;
 	register char *rc;
 
-	/* 
-	   unsigned int count = 0;
-
-	   c = arg;
-	   for(i = strlen(arg); i != 0; --i) {
-	   if(*c=='"' || *c=='\\') count++;
-	   c++;
-	   }
-	   ret = malloc(strlen(arg)+count+1);
-	   */
 	/* instead of counting in that loop above, just
 	 * use a bit more memory and half running time
 	 */
@@ -130,46 +273,37 @@ static char * mpd_sanitizeArg(const char * arg) {
 	return ret;
 }
 
-/* DONT USE THIS FUNCTION UNLESS YOU'RE ABSOLUTELY SURE WHAT YOU'RE DOING !!!! */
-static void mpd_setReturnElement(mpd_Connection * connection, const char * name, const char * value)
+static mpd_ReturnElement * mpd_newReturnElement(const char * name, const char * value)
 {
-	if (!connection) return;
-	if (!connection->returnElement) {
-		connection->returnElement = malloc(sizeof(mpd_ReturnElement));
-		connection->returnElement->name = malloc(strlen(value)+strlen(name)+3);
-	} else {
-		if ((strlen(name)+strlen(value)) > (strlen(connection->returnElement->name)+strlen(connection->returnElement->value))) {
-			free(connection->returnElement->name);
-			connection->returnElement->name = malloc(strlen(value)+strlen(name)+3);
-		}
-	}
-	memcpy(connection->returnElement->name, name, strlen(name)+strlen(value)+3);
-	connection->returnElement->value = connection->returnElement->name + strlen(name) + 2;
+	mpd_ReturnElement * ret = malloc(sizeof(mpd_ReturnElement));
+
+	ret->name = strdup(name);
+	ret->value = strdup(value);
+
+	return ret;
 }
 
-static void mpd_freeReturnElement(mpd_Connection * connection) {
-	if (connection->returnElement) {
-		free(connection->returnElement->name);
-		free(connection->returnElement);
-		connection->returnElement = NULL;
-	}
+static void mpd_freeReturnElement(mpd_ReturnElement * re) {
+	free(re->name);
+	free(re->value);
+	free(re);
 }
 
 void mpd_setConnectionTimeout(mpd_Connection * connection, float timeout) {
 	connection->timeout.tv_sec = (int)timeout;
 	connection->timeout.tv_usec = (int)(timeout*1e6 -
-			connection->timeout.tv_sec*1000000+0.5);
+	                                    connection->timeout.tv_sec*1000000 +
+					    0.5);
 }
 
 static int mpd_parseWelcome(mpd_Connection * connection, const char * host, int port,
-		char * rt, char * output) {
-
+                            char * rt, char * output) {
 	char * tmp;
 	char * test;
 	int i;
 
 	if(strncmp(output,MPD_WELCOME_MESSAGE,strlen(MPD_WELCOME_MESSAGE))) {
-		snprintf(connection->errorStr,MPD_BUFFER_MAX_LENGTH,
+		snprintf(connection->errorStr,MPD_ERRORSTR_MAX_LENGTH,
 				"mpd not running on port %i on host \"%s\"",
 				port,host);
 		connection->error = MPD_ERROR_NOTMPD;
@@ -183,10 +317,10 @@ static int mpd_parseWelcome(mpd_Connection * connection, const char * host, int
 
 		if (!tmp || (test[0] != '.' && test[0] != '\0')) {
 			snprintf(connection->errorStr,
-					MPD_BUFFER_MAX_LENGTH,
-					"error parsing version number at "
-					"\"%s\"",
-					&output[strlen(MPD_WELCOME_MESSAGE)]);
+			         MPD_ERRORSTR_MAX_LENGTH,
+			         "error parsing version number at "
+			         "\"%s\"",
+			         &output[strlen(MPD_WELCOME_MESSAGE)]);
 			connection->error = MPD_ERROR_NOTMPD;
 			return 1;
 		}
@@ -198,22 +332,11 @@ static int mpd_parseWelcome(mpd_Connection * connection, const char * host, int
 
 mpd_Connection * mpd_newConnection(const char * host, int port, float timeout) {
 	int err;
-	struct hostent * he;
-	struct sockaddr * dest;
-#ifdef HAVE_SOCKLEN_T
-	socklen_t destlen;
-#else
-	int destlen;
-#endif
-	struct sockaddr_in sin;
 	char * rt;
-	char * output;
+	char * output =  NULL;
 	mpd_Connection * connection = malloc(sizeof(mpd_Connection));
 	struct timeval tv;
 	fd_set fds;
-#ifdef MPD_HAVE_IPV6
-	struct sockaddr_in6 sin6;
-#endif
 	strcpy(connection->buffer,"");
 	connection->buflen = 0;
 	connection->bufstart = 0;
@@ -225,92 +348,12 @@ mpd_Connection * mpd_newConnection(const char * host, int port, float timeout) {
 	connection->doneListOk = 0;
 	connection->returnElement = NULL;
 	connection->request = NULL;
-#ifdef WIN32
-	WSADATA wsaData;
-	if ((WSAStartup(MAKEWORD(2, 2), &wsaData)) != 0 ||
-			LOBYTE(wsaData.wVersion) != 2 || HIBYTE(wsaData.wVersion) != 2 ) {
-		snprintf(connection->errorStr,MPD_BUFFER_MAX_LENGTH,
-				"Could not find usable WinSock DLL.");
-		connection->error = MPD_ERROR_SYSTEM;
+
+	if (winsock_dll_error(connection))
 		return connection;
-	}
-#endif
 
-	if(!(he=gethostbyname(host))) {
-		snprintf(connection->errorStr,MPD_BUFFER_MAX_LENGTH,
-				"host \"%s\" not found",host);
-		connection->error = MPD_ERROR_UNKHOST;
+	if (mpd_connect(connection, host, port, timeout) < 0)
 		return connection;
-	}
-
-	memset(&sin,0,sizeof(struct sockaddr_in));
-	/*dest.sin_family = he->h_addrtype;*/
-	sin.sin_family = AF_INET;
-	sin.sin_port = htons(port);
-#ifdef MPD_HAVE_IPV6
-	memset(&sin6,0,sizeof(struct sockaddr_in6));
-	sin6.sin6_family = AF_INET6;
-	sin6.sin6_port = htons(port);
-#endif
-	switch(he->h_addrtype) {
-		case AF_INET:
-			memcpy((char *)&sin.sin_addr.s_addr,(char *)he->h_addr,
-					he->h_length);
-			dest = (struct sockaddr *)&sin;
-			destlen = sizeof(struct sockaddr_in);
-			break;
-#ifdef MPD_HAVE_IPV6
-		case AF_INET6:
-			if(!mpd_ipv6Supported()) {
-				strcpy(connection->errorStr,"no IPv6 suuport but a "
-						"IPv6 address found\n");
-				connection->error = MPD_ERROR_SYSTEM;
-				return connection;
-			}
-			memcpy((char *)&sin6.sin6_addr.s6_addr,(char *)he->h_addr,
-					he->h_length);
-			dest = (struct sockaddr *)&sin6;
-			destlen = sizeof(struct sockaddr_in6);
-			break;
-#endif
-		default:
-			strcpy(connection->errorStr,"address type is not IPv4 or "
-					"IPv6\n");
-			connection->error = MPD_ERROR_SYSTEM;
-			return connection;
-			break;
-	}
-
-	if((connection->sock = socket(dest->sa_family,SOCK_STREAM,0))<0) {
-		strcpy(connection->errorStr,"problems creating socket");
-		connection->error = MPD_ERROR_SYSTEM;
-		return connection;
-	}
-
-	mpd_setConnectionTimeout(connection,timeout);
-
-	/* connect stuff */
-	{
-#ifdef WIN32
-		int iMode = 1; // 0 = blocking, else non-blocking
-		ioctlsocket(connection->sock, FIONBIO, (u_long FAR*) &iMode);
-		if(connect(connection->sock,dest,destlen) == SOCKET_ERROR
-				&& WSAGetLastError() != WSAEWOULDBLOCK)
-#else
-			int flags = fcntl(connection->sock, F_GETFL, 0);
-		fcntl(connection->sock, F_SETFL, flags | O_NONBLOCK);
-
-		if(connect(connection->sock,dest,destlen)<0 && 
-				errno!=EINPROGRESS) 
-#endif
-		{
-			snprintf(connection->errorStr,MPD_BUFFER_MAX_LENGTH,
-					"problems connecting to \"%s\" on port"
-					" %i",host,port);
-			connection->error = MPD_ERROR_CONNPORT;
-			return connection;
-		}
-	}
 
 	while(!(rt = strstr(connection->buffer,"\n"))) {
 		tv.tv_sec = connection->timeout.tv_sec;
@@ -323,7 +366,7 @@ mpd_Connection * mpd_newConnection(const char * host, int port, float timeout) {
 					&(connection->buffer[connection->buflen]),
 					MPD_BUFFER_MAX_LENGTH-connection->buflen,0);
 			if(readed<=0) {
-				snprintf(connection->errorStr,MPD_BUFFER_MAX_LENGTH,
+				snprintf(connection->errorStr,MPD_ERRORSTR_MAX_LENGTH,
 						"problems getting a response from"
 						" \"%s\" on port %i : %s",host,
 						port, strerror(errno));
@@ -334,20 +377,17 @@ mpd_Connection * mpd_newConnection(const char * host, int port, float timeout) {
 			connection->buffer[connection->buflen] = '\0';
 		}
 		else if(err<0) {
-			switch(errno) {
-				case EINTR:
-					continue;
-				default:
-					snprintf(connection->errorStr,
-							MPD_BUFFER_MAX_LENGTH,
-							"problems connecting to \"%s\" on port"
-							" %i",host,port);
-					connection->error = MPD_ERROR_CONNPORT;
-					return connection;
-			}
+ 			if (SELECT_ERRNO_IGNORE)
+				continue;
+			snprintf(connection->errorStr,
+					MPD_ERRORSTR_MAX_LENGTH,
+					"problems connecting to \"%s\" on port"
+					" %i",host,port);
+			connection->error = MPD_ERROR_CONNPORT;
+			return connection;
 		}
 		else {
-			snprintf(connection->errorStr,MPD_BUFFER_MAX_LENGTH,
+			snprintf(connection->errorStr,MPD_ERRORSTR_MAX_LENGTH,
 					"timeout in attempting to get a response from"
 					" \"%s\" on port %i",host,port);
 			connection->error = MPD_ERROR_NORESPONSE;
@@ -373,17 +413,11 @@ void mpd_clearError(mpd_Connection * connection) {
 }
 
 void mpd_closeConnection(mpd_Connection * connection) {
-#ifdef WIN32
 	closesocket(connection->sock);
-#else
-	close(connection->sock);
-#endif
 	if(connection->returnElement) free(connection->returnElement);
 	if(connection->request) free(connection->request);
 	free(connection);
-#ifdef WIN32
 	WSACleanup();
-#endif
 }
 
 static void mpd_executeCommand(mpd_Connection * connection, char * command) {
@@ -406,19 +440,14 @@ static void mpd_executeCommand(mpd_Connection * connection, char * command) {
 	tv.tv_sec = connection->timeout.tv_sec;
 	tv.tv_usec = connection->timeout.tv_usec;
 
-	while((ret = select(connection->sock+1,NULL,&fds,NULL,&tv)==1) || 
-			(ret==-1 && errno==EINTR)) {
-		ret = send(connection->sock,commandPtr,commandLen,
-#ifdef WIN32
-				0);
-#else
-		MSG_DONTWAIT);
-#endif
+	while((ret = select(connection->sock+1,NULL,&fds,NULL,&tv)==1) ||
+			(ret==-1 && SELECT_ERRNO_IGNORE)) {
+		ret = send(connection->sock,commandPtr,commandLen,MSG_DONTWAIT);
 		if(ret<=0)
 		{
-			if(ret==EAGAIN || ret==EINTR) continue;
-			snprintf(connection->errorStr,MPD_BUFFER_MAX_LENGTH,
-					"problems giving command \"%s\"",command);
+			if (SENDRECV_ERRNO_IGNORE) continue;
+			snprintf(connection->errorStr,MPD_ERRORSTR_MAX_LENGTH,
+			         "problems giving command \"%s\"",command);
 			connection->error = MPD_ERROR_SENDING;
 			return;
 		}
@@ -432,8 +461,8 @@ static void mpd_executeCommand(mpd_Connection * connection, char * command) {
 
 	if(commandLen>0) {
 		perror("");
-		snprintf(connection->errorStr,MPD_BUFFER_MAX_LENGTH,
-				"timeout sending command \"%s\"",command);
+		snprintf(connection->errorStr,MPD_ERRORSTR_MAX_LENGTH,
+		         "timeout sending command \"%s\"",command);
 		connection->error = MPD_ERROR_TIMEOUT;
 		return;
 	}
@@ -457,18 +486,19 @@ static void mpd_getNextReturnElement(mpd_Connection * connection) {
 	int err;
 	int pos;
 
+	if(connection->returnElement) mpd_freeReturnElement(connection->returnElement);
+	connection->returnElement = NULL;
 
 	if(connection->doneProcessing || (connection->listOks &&
-				connection->doneListOk)) 
+	   connection->doneListOk))
 	{
 		strcpy(connection->errorStr,"already done processing current command");
 		connection->error = 1;
-		mpd_freeReturnElement(connection);
 		return;
 	}
 
 	bufferCheck = connection->buffer+connection->bufstart;
-	while(connection->bufstart>=connection->buflen || 
+	while(connection->bufstart>=connection->buflen ||
 			!(rt = strchr(bufferCheck,'\n'))) {
 		if(connection->buflen>=MPD_BUFFER_MAX_LENGTH) {
 			memmove(connection->buffer,
@@ -484,7 +514,6 @@ static void mpd_getNextReturnElement(mpd_Connection * connection) {
 			connection->error = MPD_ERROR_BUFFEROVERRUN;
 			connection->doneProcessing = 1;
 			connection->doneListOk = 0;
-			mpd_freeReturnElement(connection);
 			return;
 		}
 		bufferCheck = connection->buffer+connection->buflen;
@@ -496,33 +525,27 @@ static void mpd_getNextReturnElement(mpd_Connection * connection) {
 			readed = recv(connection->sock,
 					connection->buffer+connection->buflen,
 					MPD_BUFFER_MAX_LENGTH-connection->buflen,
-#ifdef WIN32
-					0);
-#else
-			MSG_DONTWAIT);
-#endif
-			if(readed<0 && (errno==EAGAIN || errno==EINTR)) {
+					MSG_DONTWAIT);
+			if(readed<0 && SENDRECV_ERRNO_IGNORE) {
 				continue;
 			}
 			if(readed<=0) {
 				strcpy(connection->errorStr,"connection"
-						" closed");
+				       " closed");
 				connection->error = MPD_ERROR_CONNCLOSED;
 				connection->doneProcessing = 1;
 				connection->doneListOk = 0;
-				mpd_freeReturnElement(connection);
 				return;
 			}
 			connection->buflen+=readed;
 			connection->buffer[connection->buflen] = '\0';
 		}
-		else if(err<0 && errno==EINTR) continue;
+		else if(err<0 && SELECT_ERRNO_IGNORE) continue;
 		else {
 			strcpy(connection->errorStr,"connection timeout");
 			connection->error = MPD_ERROR_TIMEOUT;
 			connection->doneProcessing = 1;
 			connection->doneListOk = 0;
-			mpd_freeReturnElement(connection);
 			return;
 		}
 	}
@@ -539,13 +562,12 @@ static void mpd_getNextReturnElement(mpd_Connection * connection) {
 		connection->listOks = 0;
 		connection->doneProcessing = 1;
 		connection->doneListOk = 0;
-		mpd_freeReturnElement(connection);
 		return;
 	}
 
 	if(strcmp(output, "list_OK") == 0) {
 		if(!connection->listOks) {
-			strcpy(connection->errorStr, 
+			strcpy(connection->errorStr,
 					"got an unexpected list_OK");
 			connection->error = 1;
 		}
@@ -553,7 +575,6 @@ static void mpd_getNextReturnElement(mpd_Connection * connection) {
 			connection->doneListOk = 1;
 			connection->listOks--;
 		}
-		mpd_freeReturnElement(connection);
 		return;
 	}
 
@@ -577,34 +598,24 @@ static void mpd_getNextReturnElement(mpd_Connection * connection) {
 		val = strtol(test+1, &test, 10);
 		if(*test != ']') return;
 		connection->errorAt = val;
-		mpd_freeReturnElement(connection);
 		return;
 	}
 
 	tok = strchr(output, ':');
 	if (!tok) return;
 	pos = tok - output;
-	tok[0]='\0';
-	value = tok+1;
-	//name = (char *)malloc(pos+1);
-	//strncpy(name, output, pos);
-	//name[pos] = '\0';
-	if(value && value[0]==' ') {
-		mpd_setReturnElement(connection, output, &(value[1]));
+	value = ++tok;
+	name = output;
+	name[pos] = '\0';
+
+	if(value[0]==' ') {
+		connection->returnElement = mpd_newReturnElement(name,&(value[1]));
 	}
 	else {
-		if(!value) {
-			tok[0]=':';
-			snprintf(connection->errorStr,MPD_BUFFER_MAX_LENGTH,
-					"error parsing: %s",output);
-		}
-		else {
-			snprintf(connection->errorStr,MPD_BUFFER_MAX_LENGTH,
+		snprintf(connection->errorStr,MPD_ERRORSTR_MAX_LENGTH,
 					"error parsing: %s:%s",name,value);
-		}
-		connection->errorStr[MPD_BUFFER_MAX_LENGTH] = '\0';
+		connection->errorStr[MPD_ERRORSTR_MAX_LENGTH] = '\0';
 		connection->error = 1;
-		mpd_freeReturnElement(connection);
 	}
 }
 
@@ -616,8 +627,8 @@ void mpd_finishCommand(mpd_Connection * connection) {
 }
 
 static void mpd_finishListOkCommand(mpd_Connection * connection) {
-	while(!connection->doneProcessing && connection->listOks && 
-			!connection->doneListOk ) 
+	while(!connection->doneProcessing && connection->listOks &&
+			!connection->doneListOk)
 	{
 		mpd_getNextReturnElement(connection);
 	}
@@ -637,8 +648,12 @@ void mpd_sendStatusCommand(mpd_Connection * connection) {
 mpd_Status * mpd_getStatus(mpd_Connection * connection) {
 	mpd_Status * status;
 
-	if(connection->doneProcessing || (connection->listOks && 
-				connection->doneListOk))
+	/*mpd_executeCommand(connection,"status\n");
+
+	if(connection->error) return NULL;*/
+
+	if(connection->doneProcessing || (connection->listOks &&
+	   connection->doneListOk))
 	{
 		return NULL;
 	}
@@ -709,14 +724,13 @@ mpd_Status * mpd_getStatus(mpd_Connection * connection) {
 			status->songid = atoi(re->value);
 		}
 		else if(strcmp(re->name,"time")==0) {
-			char * copy;
-			char * temp;
-			copy = strdup(re->value);
-			status->elapsedTime = atoi(copy);
-			temp = strchr(copy, ':');
-			if (temp)
-				status->totalTime = atoi(++temp);
-			free(copy);
+			char * tok = strchr(re->value,':');
+			/* the second strchr below is a safety check */
+			if (tok && (strchr(tok,0) > (tok+1))) {
+				/* atoi stops at the first non-[0-9] char: */
+				status->elapsedTime = atoi(re->value);
+				status->totalTime = atoi(tok+1);
+			}
 		}
 		else if(strcmp(re->name,"error")==0) {
 			status->error = strdup(re->value);
@@ -728,19 +742,14 @@ mpd_Status * mpd_getStatus(mpd_Connection * connection) {
 			status->updatingDb = atoi(re->value);
 		}
 		else if(strcmp(re->name,"audio")==0) {
-			char * copy;
-			char * temp;
-			copy = strdup(re->value);
-			status->sampleRate = atoi(copy);
-			temp = strchr(copy,':');
-			if(temp) {
-				status->bits = atoi(++temp);
-				temp = strchr(temp,':');
-				if(temp) {
-					status->channels = atoi(++temp);
-				}
+			char * tok = strchr(re->value,':');
+			if (tok && (strchr(tok,0) > (tok+1))) {
+				status->sampleRate = atoi(re->value);
+				status->bits = atoi(++tok);
+				tok = strchr(tok,':');
+				if (tok && (strchr(tok,0) > (tok+1)))
+					status->channels = atoi(tok+1);
 			}
-			free(copy);
 		}
 
 		mpd_getNextReturnElement(connection);
@@ -778,10 +787,10 @@ mpd_Stats * mpd_getStats(mpd_Connection * connection) {
 
 	/*mpd_executeCommand(connection,"stats\n");
 
-	  if(connection->error) return NULL;*/
+	if(connection->error) return NULL;*/
 
-	if(connection->doneProcessing || (connection->listOks && 
-				connection->doneListOk))
+	if(connection->doneProcessing || (connection->listOks &&
+	   connection->doneListOk))
 	{
 		return NULL;
 	}
@@ -855,6 +864,9 @@ static void mpd_initSong(mpd_Song * song) {
 	/* added by Qball */
 	song->genre = NULL;
 	song->composer = NULL;
+	song->performer = NULL;
+	song->disc = NULL;
+	song->comment = NULL;
 
 	song->time = MPD_SONG_NO_TIME;
 	song->pos = MPD_SONG_NO_NUM;
@@ -871,9 +883,12 @@ static void mpd_finishSong(mpd_Song * song) {
 	if(song->date) free(song->date);
 	if(song->genre) free(song->genre);
 	if(song->composer) free(song->composer);
+	if(song->performer) free(song->performer);
+	if(song->disc) free(song->disc);
+	if(song->comment) free(song->comment);
 }
 
-mpd_Song * mpd_newSong() {
+mpd_Song * mpd_newSong(void) {
 	mpd_Song * ret = malloc(sizeof(mpd_Song));
 
 	mpd_initSong(ret);
@@ -898,6 +913,9 @@ mpd_Song * mpd_songDup(mpd_Song * song) {
 	if(song->date) ret->date = strdup(song->date);
 	if(song->genre) ret->genre= strdup(song->genre);
 	if(song->composer) ret->composer= strdup(song->composer);
+	if(song->performer) ret->performer= strdup(song->performer);
+	if(song->disc) ret->disc = strdup(song->disc);
+	if(song->comment) ret->comment = strdup(song->comment);
 	ret->time = song->time;
 	ret->pos = song->pos;
 	ret->id = song->id;
@@ -913,7 +931,7 @@ static void mpd_finishDirectory(mpd_Directory * directory) {
 	if(directory->path) free(directory->path);
 }
 
-mpd_Directory * mpd_newDirectory () {
+mpd_Directory * mpd_newDirectory(void) {
 	mpd_Directory * directory = malloc(sizeof(mpd_Directory));;
 
 	mpd_initDirectory(directory);
@@ -943,7 +961,7 @@ static void mpd_finishPlaylistFile(mpd_PlaylistFile * playlist) {
 	if(playlist->path) free(playlist->path);
 }
 
-mpd_PlaylistFile * mpd_newPlaylistFile() {
+mpd_PlaylistFile * mpd_newPlaylistFile(void) {
 	mpd_PlaylistFile * playlist = malloc(sizeof(mpd_PlaylistFile));
 
 	mpd_initPlaylistFile(playlist);
@@ -966,7 +984,7 @@ mpd_PlaylistFile * mpd_playlistFileDup(mpd_PlaylistFile * playlist) {
 
 static void mpd_initInfoEntity(mpd_InfoEntity * entity) {
 	entity->info.directory = NULL;
-} 
+}
 
 static void mpd_finishInfoEntity(mpd_InfoEntity * entity) {
 	if(entity->info.directory) {
@@ -982,7 +1000,7 @@ static void mpd_finishInfoEntity(mpd_InfoEntity * entity) {
 	}
 }
 
-mpd_InfoEntity * mpd_newInfoEntity() {
+mpd_InfoEntity * mpd_newInfoEntity(void) {
 	mpd_InfoEntity * entity = malloc(sizeof(mpd_InfoEntity));
 
 	mpd_initInfoEntity(entity);
@@ -1002,20 +1020,20 @@ static void mpd_sendInfoCommand(mpd_Connection * connection, char * command) {
 mpd_InfoEntity * mpd_getNextInfoEntity(mpd_Connection * connection) {
 	mpd_InfoEntity * entity = NULL;
 
-	if(connection->doneProcessing || (connection->listOks && 
-				connection->doneListOk))
+	if(connection->doneProcessing || (connection->listOks &&
+	   connection->doneListOk))
 	{
 		return NULL;
 	}
 
 	if(!connection->returnElement) mpd_getNextReturnElement(connection);
 
-	if(connection->returnElement) { 
+	if(connection->returnElement) {
 		if(strcmp(connection->returnElement->name,"file")==0) {
 			entity = mpd_newInfoEntity();
 			entity->type = MPD_INFO_ENTITY_TYPE_SONG;
 			entity->info.song = mpd_newSong();
-			entity->info.song->file = 
+			entity->info.song->file =
 				strdup(connection->returnElement->value);
 		}
 		else if(strcmp(connection->returnElement->name,
@@ -1023,16 +1041,22 @@ mpd_InfoEntity * mpd_getNextInfoEntity(mpd_Connection * connection) {
 			entity = mpd_newInfoEntity();
 			entity->type = MPD_INFO_ENTITY_TYPE_DIRECTORY;
 			entity->info.directory = mpd_newDirectory();
-			entity->info.directory->path = 
+			entity->info.directory->path =
 				strdup(connection->returnElement->value);
 		}
 		else if(strcmp(connection->returnElement->name,"playlist")==0) {
 			entity = mpd_newInfoEntity();
 			entity->type = MPD_INFO_ENTITY_TYPE_PLAYLISTFILE;
 			entity->info.playlistFile = mpd_newPlaylistFile();
-			entity->info.playlistFile->path = 
+			entity->info.playlistFile->path =
 				strdup(connection->returnElement->value);
 		}
+		else if(strcmp(connection->returnElement->name, "cpos") == 0){
+			entity = mpd_newInfoEntity();
+			entity->type = MPD_INFO_ENTITY_TYPE_SONG;
+			entity->info.song = mpd_newSong();
+			entity->info.song->pos = atoi(connection->returnElement->value);
+		}
 		else {
 			connection->error = 1;
 			strcpy(connection->errorStr,"problem parsing song info");
@@ -1048,8 +1072,9 @@ mpd_InfoEntity * mpd_getNextInfoEntity(mpd_Connection * connection) {
 		if(strcmp(re->name,"file")==0) return entity;
 		else if(strcmp(re->name,"directory")==0) return entity;
 		else if(strcmp(re->name,"playlist")==0) return entity;
+		else if(strcmp(re->name,"cpos")==0) return entity;
 
-		if(entity->type == MPD_INFO_ENTITY_TYPE_SONG && 
+		if(entity->type == MPD_INFO_ENTITY_TYPE_SONG &&
 				strlen(re->value)) {
 			if(!entity->info.song->artist &&
 					strcmp(re->name,"Artist")==0) {
@@ -1094,8 +1119,19 @@ mpd_InfoEntity * mpd_getNextInfoEntity(mpd_Connection * connection) {
 			else if(!entity->info.song->composer &&
 					strcmp(re->name, "Composer") == 0) {
 				entity->info.song->composer = strdup(re->value);
-			}                                                    			
-
+			}
+			else if(!entity->info.song->performer &&
+					strcmp(re->name, "Performer") == 0) {
+				entity->info.song->performer = strdup(re->value);
+			}
+			else if(!entity->info.song->disc &&
+					strcmp(re->name, "Disc") == 0) {
+				entity->info.song->disc = strdup(re->value);
+			}
+			else if(!entity->info.song->comment &&
+					strcmp(re->name, "Comment") == 0) {
+				entity->info.song->comment = strdup(re->value);
+			}
 		}
 		else if(entity->type == MPD_INFO_ENTITY_TYPE_DIRECTORY) {
 		}
@@ -1108,11 +1144,11 @@ mpd_InfoEntity * mpd_getNextInfoEntity(mpd_Connection * connection) {
 	return entity;
 }
 
-static char * mpd_getNextReturnElementNamed(mpd_Connection * connection, 
-		const char * name) 
+static char * mpd_getNextReturnElementNamed(mpd_Connection * connection,
+		const char * name)
 {
-	if(connection->doneProcessing || (connection->listOks && 
-				connection->doneListOk)) 
+	if(connection->doneProcessing || (connection->listOks &&
+				connection->doneListOk))
 	{
 		return NULL;
 	}
@@ -1128,13 +1164,16 @@ static char * mpd_getNextReturnElementNamed(mpd_Connection * connection,
 	return NULL;
 }
 
-	char * mpd_getNextTag(mpd_Connection * connection,int table) {
-		if(table >= 0 && table < MPD_TAG_NUM_OF_ITEM_TYPES)
-		{
-			return mpd_getNextReturnElementNamed(connection,mpdTagItemKeys[table]);
-		}
+char *mpd_getNextTag(mpd_Connection *connection, int type)
+{
+	if (type < 0 || type >= MPD_TAG_NUM_OF_ITEM_TYPES ||
+	    type == MPD_TAG_ITEM_ANY)
 		return NULL;
-	}
+	if (type == MPD_TAG_ITEM_FILENAME)
+		return mpd_getNextReturnElementNamed(connection, "file");
+	return mpd_getNextReturnElementNamed(connection, mpdTagItemKeys[type]);
+}
+
 char * mpd_getNextArtist(mpd_Connection * connection) {
 	return mpd_getNextReturnElementNamed(connection,"Artist");
 }
@@ -1164,6 +1203,13 @@ void mpd_sendPlChangesCommand(mpd_Connection * connection, long long playlist) {
 	free(string);
 }
 
+void mpd_sendPlChangesPosIdCommand(mpd_Connection * connection, long long playlist) {
+	char * string = malloc(strlen("plchangesposid")+25);
+	sprintf(string,"plchangesposid \"%lld\"\n",playlist);
+	mpd_sendInfoCommand(connection,string);
+	free(string);
+}
+
 void mpd_sendListallCommand(mpd_Connection * connection, const char * dir) {
 	char * sDir = mpd_sanitizeArg(dir);
 	char * string = malloc(strlen("listall")+strlen(sDir)+5);
@@ -1195,8 +1241,8 @@ void mpd_sendCurrentSongCommand(mpd_Connection * connection) {
 	mpd_executeCommand(connection,"currentsong\n");
 }
 
-void mpd_sendSearchCommand(mpd_Connection * connection, int table, 
-		const char * str) 
+void mpd_sendSearchCommand(mpd_Connection * connection, int table,
+		const char * str)
 {
 	char st[10];
 	char * string;
@@ -1217,8 +1263,8 @@ void mpd_sendSearchCommand(mpd_Connection * connection, int table,
 	free(sanitStr);
 }
 
-void mpd_sendFindCommand(mpd_Connection * connection, int table, 
-		const char * str) 
+void mpd_sendFindCommand(mpd_Connection * connection, int table,
+		const char * str)
 {
 	char st[10];
 	char * string;
@@ -1238,8 +1284,8 @@ void mpd_sendFindCommand(mpd_Connection * connection, int table,
 	free(sanitStr);
 }
 
-void mpd_sendListCommand(mpd_Connection * connection, int table, 
-		const char * arg1) 
+void mpd_sendListCommand(mpd_Connection * connection, int table,
+		const char * arg1)
 {
 	char st[10];
 	char * string;
@@ -1264,7 +1310,6 @@ void mpd_sendListCommand(mpd_Connection * connection, int table,
 	free(string);
 }
 
-
 void mpd_sendAddCommand(mpd_Connection * connection, const char * file) {
 	char * sFile = mpd_sanitizeArg(file);
 	char * string = malloc(strlen("add")+strlen(sFile)+5);
@@ -1274,6 +1319,26 @@ void mpd_sendAddCommand(mpd_Connection * connection, const char * file) {
 	free(sFile);
 }
 
+int mpd_sendAddIdCommand(mpd_Connection *connection, const char *file)
+{
+	int retval = -1;
+	char *sFile = mpd_sanitizeArg(file);
+	char *string = malloc(strlen("addid")+strlen(sFile)+5);
+
+	sprintf(string, "addid \"%s\"\n", sFile);
+	mpd_sendInfoCommand(connection, string);
+	free(string);
+	free(sFile);
+
+	string = mpd_getNextReturnElementNamed(connection, "Id");
+	if (string) {
+		retval = atoi(string);
+		free(string);
+	}
+	
+	return retval;
+}
+
 void mpd_sendDeleteCommand(mpd_Connection * connection, int songPos) {
 	char * string = malloc(strlen("delete")+25);
 	sprintf(string,"delete \"%i\"\n",songPos);
@@ -1560,8 +1625,6 @@ void mpd_freeOutputElement(mpd_OutputEntity * output) {
 	free(output);
 }
 
-
-
 /**
  * mpd_sendNotCommandsCommand
  * odd naming, but it gets the not allowed commands
@@ -1579,105 +1642,137 @@ void mpd_sendNotCommandsCommand(mpd_Connection * connection) {
 void mpd_sendCommandsCommand(mpd_Connection * connection) {
 	mpd_executeCommand(connection,"commands\n");
 }
-
+/**
+ * Get the next returned command
+ */
 char * mpd_getNextCommand(mpd_Connection * connection) {
 	return mpd_getNextReturnElementNamed(connection,"command");
 }
 
-void mpd_startSearch(mpd_Connection * connection,int exact) {
-	if(connection->request) {
-		/* search/find allready in progress */
-		/* TODO: set error here?  */
-		return;
-	}
-	if(exact){
-		connection->request = strdup("find");
-	}
-	else{
-		connection->request = strdup("search");
-	}
-}
-
-
-void mpd_startFieldSearch(mpd_Connection * connection,int field) {
-	if(connection->request) {
-		/* search/find allready in progress */
-		/* TODO: set error here?  */
-		return;
-	}
-	if(field < 0 || field >= MPD_TAG_NUM_OF_ITEM_TYPES) {
-		/* set error here */
-		return;
-	}
-
-	connection->request = malloc(sizeof(char)*(
-				/* length of the field name */
-				strlen(mpdTagItemKeys[field])+ 	
-				/* "list"+space+\0 */
-				6 
-				));
-	sprintf(connection->request, "list %s", mpdTagItemKeys[field]);
-}
-
-
-
-void mpd_addConstraintSearch(mpd_Connection *connection, 
-		int field, 
-		char *name)
+void mpd_startSearch(mpd_Connection *connection, int exact)
 {
-	char *arg = NULL;
-	if(!connection->request){
+	if (connection->request) {
+		strcpy(connection->errorStr, "search already in progress");
+		connection->error = 1;
 		return;
 	}
-	if(name == NULL) {
+
+	if (exact) connection->request = strdup("find");
+	else connection->request = strdup("search");
+}
+
+void mpd_startFieldSearch(mpd_Connection *connection, int type)
+{
+	char *strtype;
+
+	if (connection->request) {
+		strcpy(connection->errorStr, "search already in progress");
+		connection->error = 1;
 		return;
 	}
-	if(field < 0 || field >= MPD_TAG_NUM_OF_ITEM_TYPES) {
+
+	if (type < 0 || type >= MPD_TAG_NUM_OF_ITEM_TYPES) {
+		strcpy(connection->errorStr, "invalid type specified");
+		connection->error = 1;
 		return;
 	}
-	/* clean up the query */
+
+	strtype = mpdTagItemKeys[type];
+
+	connection->request = malloc(strlen(strtype)+6 /* "list"+space+\0 */);
+
+	sprintf(connection->request, "list %c%s",
+	        (char)tolower(strtype[0]), strtype+1);
+}
+
+void mpd_addConstraintSearch(mpd_Connection *connection, int type, char *name)
+{
+	char *strtype;
+	char *arg;
+
+	if (!connection->request) {
+		strcpy(connection->errorStr, "no search in progress");
+		connection->error = 1;
+		return;
+	}
+
+	if (type < 0 || type >= MPD_TAG_NUM_OF_ITEM_TYPES) {
+		strcpy(connection->errorStr, "invalid type specified");
+		connection->error = 1;
+		return;
+	}
+
+	if (name == NULL) {
+		strcpy(connection->errorStr, "no name specified");
+		connection->error = 1;
+		return;
+	}
+
+	strtype = mpdTagItemKeys[type];
 	arg = mpd_sanitizeArg(name);
-	/* create space for the query */
-	connection->request = realloc(connection->request, (
-			 /* length of the old string */
-			 strlen(connection->request)+ 
-			 /* space between */
-			 1+
-			 /* length of the field name */
-			 strlen(mpdTagItemKeys[field])+ 
-			 /* space plus starting " */
-			 2+
-			 /* length of search term */
-			 strlen(arg)+
-			 /* closign " +\0 that is added sprintf */
-			 2
-			)*sizeof(char));
-	/* and form the query */
-	sprintf(connection->request, "%s %s \"%s\"",
-			connection->request,
-			mpdTagItemKeys[field],
-			arg);
+
+	connection->request = realloc(connection->request,
+	                              strlen(connection->request)+
+	                              strlen(strtype)+strlen(arg)+
+	                              5 /* two spaces+two quotes+\0 */);
+
+	sprintf(connection->request, "%s %c%s \"%s\"", connection->request,
+	        (char)tolower(strtype[0]), strtype+1, arg);
+
 	free(arg);
 }
 
-
-void mpd_commitSearch(mpd_Connection *connection) 
+void mpd_commitSearch(mpd_Connection *connection)
 {
-	if(connection->request)
-	{
-		int length = strlen(connection->request);
-		/* fixing up the string for mpd to like */
-		connection->request = realloc(connection->request, 
-				(length+	/* old length */
-				 2		/* closing \n and \0 */
-				)*sizeof(char));
-		connection->request[length] = '\n';
-		connection->request[length+1] = '\0';
-		/* and off we go */
-		mpd_sendInfoCommand(connection, connection->request);
-		/* clean up a bit */
-		free(connection->request);
-		connection->request = NULL;
+	int length;
+
+	if (!connection->request) {
+		strcpy(connection->errorStr, "no search in progress");
+		connection->error = 1;
+		return;
 	}
+
+	length = strlen(connection->request);
+	connection->request = realloc(connection->request,
+	                              length+2 /* old length+\n+\0 */);
+	connection->request[length] = '\n';
+	connection->request[length+1] = '\0';
+	mpd_sendInfoCommand(connection, connection->request);
+
+	free(connection->request);
+	connection->request = NULL;
 }
 
+/**
+ * @param connection a MpdConnection
+ * @param path	the path to the playlist.
+ *
+ * List the content, with full metadata, of a stored playlist.
+ *
+ */
+void mpd_sendListPlaylistInfoCommand(mpd_Connection *connection, char *path)
+{
+	char *arg = mpd_sanitizeArg(path);
+	char *query = malloc(strlen("listplaylistinfo")+strlen(arg)+5);
+	sprintf(query, "listplaylistinfo \"%s\"\n",arg);
+	mpd_sendInfoCommand(connection, query);
+	free(arg);
+	free(query);
+}
+
+/**
+ * @param connection a MpdConnection
+ * @param path	the path to the playlist.
+ *
+ * List the content of a stored playlist.
+ *
+ */
+void mpd_sendListPlaylistCommand(mpd_Connection *connection, char *path)
+{
+	char *arg = mpd_sanitizeArg(path);
+	char *query = malloc(strlen("listplaylist")+strlen(arg)+5);
+	sprintf(query, "listplaylist \"%s\"\n",arg);
+	mpd_sendInfoCommand(connection, query);
+	free(arg);
+	free(query);
+}
